% $Id: breitensuche.pl,v 1.2 1995/11/29 16:11:44 uwe Exp $ 
/* Breitensuche in einem Graphen
   schon besuchte Knoten werden nicht wieder beruecksichtigt
   mit Erklaerungskomponente (Pfad)

   ein Pfad wird als Liste mit dem Operator ':->'
   und dem '+' operator aufgebaut

  si :-> ri + (si-1 :-> .... + (s0 -> nil))
*/

%-- trace output ---------------------------------------------------------------------

:- consult('trace').

%--- generiere alle Nachfolger aus einem Knoten --------------------------------------
  
:- op(680,xfx,':->').

gen1(P,Succ) :-
	P = S :-> _ ,
	findall(S1 :-> R + P,
		rule(R,S,S1),
		Succ).

%--- generiere alle Nachfolger aus einer Liste ---------------------------------------

gen([], []).

gen([S | T], Succ) :-
	gen1(S, Succ1),
	gen(T,  Succ2),
	union(Succ1, Succ2, Succ).
  
%--- union(S1,S2,S1 u S2) ------------------------------------------------------------

union([], S, S).

union([H :-> _ | T], S2, S) :-
	isin(H, S2),
	!,
	union(T, S2, S).

union([H | T], S2, [H | S]) :-
	union(T, S2, S).
    
%--- minus(S1,S2,S1 - S2) ------------------------------------------------------------

minus(S1, [], S1).

minus(S1, [H :-> _ | S2], S3) :-
	minus1(H, S1, S4),
	minus(S4, S2, S3).

minus1(_, [], []).

minus1(E, [S1 :-> _ | T], S) :-
	E = S1,
	!,
	minus1(E, T, S).

minus1(E, [H | T], [H | S]) :-
	minus1(E, T, S). 

isin(S, [H :-> _ | _]) :-
	S = H,
	!.

isin(S, [_ | T]) :-
	isin(S, T).

card([], 0) :-
	!.
card([_ | R], I) :-
	card(R, I1),
	I is I1 + 1.

%--- Loesungen filtern ---------------------------------------------------------------

solutions([], []).
solutions([H | R], [H | SL]) :-
	H = S :-> _,
	solution(S),
	!,
	solutions(R,SL).

solutions([_ | R], SL) :-
	solutions(R, SL).

%--- Generierungsprozess -------------------------------------------------------------

    % solve(+Statelist, +N, +Max, +Visited, -Result)
    % Loesung gefunden 
solve(Statelist, N, _, _, found(N, Solutions)) :-
	solutions(Statelist,Solutions),
	Solutions \= [],
	!.

    % suche abbrechen  
solve(_, Max, Max, _, limit(Max)) :-
	!.

solve([], N, _, _, empty(N)) :-
	!.

                                            % naechste Generation von Knoten erzeugen 
solve(Statelist, N, Max, Visited, Solutions) :-
	gen(Statelist, Succ),
	union(Visited, Statelist, Newvisited),
	minus(Succ, Newvisited, Newsucc),
	N1 is N + 1,
	traceout((write(N1),write('. Schritt '),
	          card(Newsucc,N2),write(N2),write(' neue Knoten generiert'),nl)),
	solve(Newsucc, N1, Max, Newvisited, Solutions).

    % solve(+Initialstate, +MaxSearchDepth, -Result) 

solve(Initial, Max, Solutions) :-
	solve([Initial :-> nil], 0, Max, [], Solutions).
	
solve(Initial, Max) :-
	write('Start:'), nl,
	outstate(Initial), nl,
	solve(Initial, Max, Solutions),
	outsolutions(Solutions).

%--- Ausgabe -------------------------------------------------------------------------

:-assert(reversepath(off)).

reverseoff :-
	retractall(reversepath(_)),
	assert(reversepath(off)).

reverseon :-
	retractall(reversepath(_)),
	assert(reversepath(on)).

outsolutions(limit(N)) :-
	!,
	write('Keine Loesung in '), write(N), write(' Suchschritten gefunden'), nl.

outsolutions(empty(N)) :-
	!,
	write('Nach '), write(N), write('Schritten Suchraum erschoepft'), nl.

outsolutions(found(N,[Solution])) :-
	!,
	write('Nach '), write(N), write(' Schritten eine Loesung gefunden'), nl,
	outsolutions([Solution]).

outsolutions(found(N, Solutions)) :-
	!,
	write('Nach '), write(N), write(' Schritten mehrere Loesungen gefunden'), nl,
	outsolutions(Solutions).

outsolutions([]).

outsolutions([P|R]) :-
        write('Loesung:'),
	nl,
	outpath(P),
	nl,
	outsolutions(R).

outpath(P) :-
	reversepath(off),
	outpath1(P).
outpath(P) :-
	reversepath(on),
	outpath2(P).

outpath1(S:->nil) :-
	outstate(S).

outpath1(S:->R+P) :-
	outpath1(P),
	outrule(R),
	outstate(S).

outpath2(S:->nil) :-
	outstate(S).

outpath2(S:->R+P) :-
	outstate(S),
	outrule(R),
	outpath2(P).

	
